import {
  ConversationType,
  logger,
  MessageDirection,
  ReceivedStatus
} from "@rongcloud/engine";
import {
  setConversationNotificationStatus,
  setConversationNotificationLevel,
  setConversationToTop,
  removeConversation,
  getConversationList
} from "@rongcloud/imlib-next";
import { LogTagId } from "./enum/logEnums";
// 填充的空会话标识
const PolyfillMessage = "polyfill";
export default class ConversationManager {
  constructor() {
    // 第一次拉取数量
    this.FIRST_COUNT_BASE = 30;
    // 每次的拉取数量
    this.COUNT_BASE = 30;
    // 是否为第一次拉取
    this.first = true;
    /**
     * 会话分为置顶会话和非置顶会话，
     */
    this.topConversations = [];
    this.unTopConversations = [];
  }
  /**
   * 设置拉取数量
   * @param option
   */
  setPullCount(option) {
    this.COUNT_BASE = option.pullCount;
    this.FIRST_COUNT_BASE = option.pullCount;
  }
  /**
   * 获取远程会话
   * @param count
   * @param startTime
   * @param order
   * @returns Promise<void | IAsyncRes<IAReceivedConversation[]>>
   */
  getRemoteConversation(count, startTime = 0, conversationStorage, order = 0) {
    return getConversationList({
      count,
      startTime,
      order
    }).then(result => {
      if (result.code === 0) {
        let data = result.data;
        data.forEach(conversation => {
          const unReadTime = conversationStorage.getUnReadTime(
            conversation.conversationType,
            conversation.targetId,
            conversation.channelId
          );
          // 根据本地存储的已读响应时间戳判断单聊消息的已读未读状态。
          const isPrivat =
            conversation.latestMessage.conversationType ===
            ConversationType.PRIVATE;
          const isSendUser =
            conversation.latestMessage.messageDirection ===
            MessageDirection.SEND;
          // 大于本地已读存储时间或者没有本地存储已读时间均设置为未读
          const isUnRead =
            conversation.latestMessage.sentTime > unReadTime || !unReadTime;
          if (isPrivat && isSendUser && isUnRead) {
            // 单聊，自己发送的消息
            conversation.latestMessage.receivedStatus = ReceivedStatus.UNREAD;
            // console.info('修改未读', conversation)
          }
        });
        // 排除用户需要自己接管的数据
        if (
          this.customIntercept &&
          typeof this.customIntercept.interceptConversation === "function"
        ) {
          data = data.filter(one => {
            const isMatch = this.customIntercept.interceptConversation(one);
            return !isMatch;
          });
        }
        // 用户处理lastMessage
        if (
          this.customDisplayMessage &&
          this.customDisplayMessage.willDisplayConversationMessage &&
          typeof this.customDisplayMessage.willDisplayConversationMessage ===
            "function"
        ) {
          data = data.map(one => {
            let lastMessage = one.latestMessage;
            if (lastMessage) {
              let content = this.customDisplayMessage.willDisplayConversationMessage(
                lastMessage.content,
                lastMessage.conversationType,
                lastMessage.targetId,
                lastMessage.messageType
              );
              one.latestMessage.content = content;
            }
            return one;
          });
        }
        this.update(
          data.map(one => {
            return { conversation: one };
          })
        );
      }
    });
  }
  /**
   * 获取一个会话
   * @param target
   * @param source
   * @returns number
   */
  findConversationFromList(target, source) {
    const index = source.findIndex(one => {
      return (
        one.conversationType === target.conversationType &&
        one.targetId === target.targetId &&
        one.channelId === target.channelId
      );
    });
    return index;
  }
  /**
   * 在本地缓存数据中找一个会话
   * @param target 目标会话参数
   * @returns IAReceivedConversation 会话
   */
  findConversation(target) {
    target.channelId = target.channelId || "";
    let index = this.findConversationFromList(target, this.topConversations);
    if (index !== -1) {
      return {
        source: this.topConversations,
        index
      };
    }
    index = this.findConversationFromList(target, this.unTopConversations);
    if (index !== -1) {
      return {
        source: this.unTopConversations,
        index
      };
    }
    return null;
  }
  /**
   * 接受SDK返回的会话数据
   * @param conversationList
   */
  update(conversationList) {
    // 找到空会话
    const polyfillTopConversation = this.topConversations.filter(one => {
      return (
        one.latestMessage && one.latestMessage.messageType === PolyfillMessage
      );
    });
    const polyfillUnTopConversation = this.unTopConversations.filter(one => {
      return (
        one.latestMessage && one.latestMessage.messageType === PolyfillMessage
      );
    });
    const polyfillConversation = polyfillTopConversation.concat(
      polyfillUnTopConversation
    );
    conversationList = conversationList.concat(
      polyfillConversation.map(one => {
        return {
          conversation: one
        };
      })
    );
    conversationList.forEach(one => {
      if (!one || (!one.conversation && !one.updatedItems)) {
        return;
      }
      // 找到会话
      const matchResult = this.findConversation({
        conversationType: one.conversation.conversationType,
        targetId: one.conversation.targetId,
        channelId: one.conversation.channelId || "" // channelId 可能会是undefined
      });
      let target;
      target = matchResult
        ? matchResult.source[matchResult.index]
        : one.conversation;
      // 新数据字段可能不完整，用新数据存在的字段更新老的会话数据
      this.merge(target, one.conversation);
      if (
        one.updatedItems !== undefined &&
        Object.keys(one.updatedItems).length
      ) {
        one.updatedItems.hasOwnProperty("isTop") &&
          matchResult &&
          matchResult.source.splice(matchResult.index, 1);
        target &&
          Object.keys(one.updatedItems).forEach(key => {
            target[key] = one.updatedItems[key].val;
          });
      } else if (one.conversation) {
        matchResult && matchResult.source.splice(matchResult.index, 1);
      }
      // logger.info(LogTagId.A_KIT_CONVERSATION_UPDATE_O, { target, conversationUpdate: one })
      if (target.isTop) {
        this.sortConversation(target, this.topConversations);
      } else {
        this.sortConversation(target, this.unTopConversations);
      }
    });
  }
  merge(target, source) {
    for (const key in source) {
      if (source[key] !== undefined && source[key] !== null) {
        target[key] = source[key];
      }
    }
  }
  /**
   * 会话排序，按照会话的最后一条消息
   * @param source
   * @param target
   * @returns void
   */
  sortConversation(source, target) {
    if (!target.length) {
      target.push(source);
    }
    // 查看是否已经存在
    const index = target.findIndex(conversation => {
      return (
        conversation.targetId === source.targetId &&
        conversation.channelId === source.channelId &&
        conversation.conversationType === source.conversationType
      );
    });
    // 如果存在则先删除
    if (index !== -1) {
      if (target[index].latestMessage) {
        if (source.latestMessage) {
          if (
            target[index].latestMessage.sentTime > source.latestMessage.sentTime
          ) {
            // 当前会话的最后一条消息时间大于新会话数据的消息时间
            return;
          }
        } else {
          // 新的会话数据没有最后一条消息
          return;
        }
      }
      // source = Object.assign({}, target[index], source)
      target.splice(index, 1);
    }
    let position = -1;
    for (let i = 0; i < target.length; i++) {
      const conversation = target[i];
      if (conversation.latestMessage) {
        if (
          source.latestMessage &&
          source.latestMessage.sentTime >= conversation.latestMessage.sentTime
        ) {
          position = i;
          break;
        } else {
          continue;
        }
      } else {
        position = i;
        break;
      }
    }
    if (position === -1) {
      position = target.length;
    }
    target.splice(position, 0, source);
  }
  /**
   * 创建一个新会话，并插入到本地会话中
   * @param option
   */
  create(option) {
    const conversation = {
      conversationType: option.conversationType,
      channelId: option.channelId,
      targetId: option.targetId,
      latestMessage: {
        conversationType: option.conversationType,
        targetId: option.targetId,
        channelId: option.channelId,
        sentTime: new Date().getTime(),
        messageType: PolyfillMessage,
        content: "",
        senderUserId: "",
        receivedTime: 0,
        messageUId: "none",
        messageDirection: 1,
        isPersited: true,
        isCounted: true,
        isOffLineMessage: false,
        canIncludeExpansion: false,
        receivedStatus: ReceivedStatus.UNREAD
      }
    };
    this.sortConversation(conversation, this.unTopConversations);
    logger.info(LogTagId.A_KIT_CONVERSATION_CREATE_O, conversation);
  }
  /**
   * 移除一个会话
   * @param conversation
   * @returns void
   */
  remove(conversation) {
    return removeConversation(conversation).then(res => {
      const result = this.findConversation(conversation);
      if (result) {
        if (res.code === 0) {
          result.source.splice(result.index, 1);
          logger.info(LogTagId.A_KIT_CONVERSATION_DELETE_O, {
            index: result.index,
            conversation: result.source[result.index]
          });
        }
      }
      return res;
    });
  }
  /**
   * 获取所有会话中最新的消息发送时间
   * @param conversationList
   * @returns number
   */
  getLastTime(conversationList) {
    let lastTime = new Date().getTime();
    for (let i = conversationList.length - 1; i >= 0; i--) {
      const conversation = conversationList[i];
      if (conversation.latestMessage) {
        lastTime = conversation.latestMessage.sentTime;
        break;
      }
    }
    return lastTime;
  }
  /**
   * 获取会话列表
   * @param more 为true的情况下从服务端拉取，如果是false则从本地获取
   * @returns Promise<IAReceivedConversation[]>
   */
  getList(more, conversationStorage) {
    if (more) {
      // if (more || this.first) {
      // 找到最后一条会话的sentTime
      const topLastTime = this.getLastTime(this.topConversations);
      const unTopLastTime = this.getLastTime(this.unTopConversations);
      const count = this.first ? this.FIRST_COUNT_BASE : this.COUNT_BASE;
      const lastTime = Math.min(topLastTime, unTopLastTime);
      return this.getRemoteConversation(
        count,
        lastTime,
        conversationStorage
      ).then(() => {
        this.first = false;
        const list = [...this.topConversations, ...this.unTopConversations];
        return list;
      });
    } else {
      const list = [...this.topConversations, ...this.unTopConversations];
      return Promise.resolve(list);
    }
  }
  /**
   * 改变会话置顶状态
   * @param option
   * @param status
   * @returns
   */
  changeTopStatus(option, status) {
    return setConversationToTop(option, status).then(res => {
      if (res.code === 0) {
        // 改变状态之后，不需要本地重置，sdk会触发conversation事件
        // 会话已经被更新
        // const result = this.findConversation(option);
        // if (status) {
        //   // 置顶操作
        //   if (result.source === this.unTopConversations) {
        //     const conversation:IAReceivedConversation = this.unTopConversations.splice(result.index, 1)[0];
        //     conversation.isTop = true;
        //     this.sortConversation(conversation, this.topConversations);
        //   }
        // } else {
        //   // 取消置顶操作
        //   if (result.source === this.topConversations) {
        //     const conversation:IAReceivedConversation = this.topConversations.splice(result.index, 1)[0];
        //     conversation.isTop = false;
        //     this.sortConversation(conversation, this.unTopConversations)
        //   }
        // }
      }
      return res;
    });
  }
  /**
   * 改变通知状态
   * @param option
   * @param status
   * @returns
   */
  changeNotificationStatus(option, status) {
    return setConversationNotificationLevel(option, status).then(res => {
      if (res.code === 0) {
        const result = this.findConversation(option);
        result.source[result.index].notificationLevel = status;
      }
      return res;
    });
  }
  /**
   * disconnect 置空会话信息
   */
  clearLocalConversation() {
    this.unTopConversations = [];
    this.topConversations = [];
  }
}
