import {
  getHistoryMessages,
  MessageDirection,
  ReceivedStatus,
  ConversationType,
} from "@rongcloud/imlib-next";
import { logger } from "@rongcloud/engine";
import { LogTagId } from "./enum/logEnums";
import { getConversationKey, isEqual } from "./utils";
export default class MessageManager {
  constructor() {
    /**
     * 最近的消息列表
     */
    this.localMessages = {};
  }
  /**
   * 设置当前会话
   */
  set currentConversation(conversation) {
    this._currentConversation = conversation;
  }
  /**
   * 获取当前会话
   */
  get currentConversation() {
    return this._currentConversation;
  }
  /**
   * 获取消息，如果本地消息不足，则从服务端拉取
   * @param conversation 会话
   * @param start 开始时间，默认从最新一条的时间开始
   * @param count 获取数量，默认20条
   * @returns
   */
  getMessages(conversation, startTime = 0, count = 20, unReadTime) {
    count = Math.min(20, Math.max(1, count));
    this._unReadTime = unReadTime;
    const result = this.getLocalMessage(conversation, startTime, count);
    // 拿到本地消息匹配消息是否已读
    result.list &&
      result.list.forEach((message) => {
        const isPrivat = message.conversationType === ConversationType.PRIVATE;
        const isSendUser = message.messageDirection === MessageDirection.SEND;
        const isUnRead = message.sentTime <= this._unReadTime;
        if (isPrivat && isSendUser && isUnRead) {
          // 单聊，自己发送的消息
          message.receivedStatus = ReceivedStatus.READ;
        }
      });
    if (result.list.length >= count) {
      // 消息数量能满足count
      return Promise.resolve(result);
    } else {
      if (result.hasMore === false) {
        // 服务器没有更多历史消息
        return Promise.resolve(result);
      } else {
        const c = getConversationKey(conversation);
        let timestamp = 0;
        if (this.localMessages[c] && this.localMessages[c].list.length > 0) {
          timestamp = this.localMessages[c].list[0].sentTime;
        }
        return this.getRemoteMessages(conversation, { count, timestamp }).then(
          () => {
            // 已经存入本地数据，从本地数据拿出
            return this.getLocalMessage(conversation, startTime, count);
          }
        );
      }
    }
  }
  /**
   * 获取本地消息
   * @param conversation 会话
   * @param startTime 从什么时间开始获取
   * @param count 获取数量
   * @returns { list: IAReceivedMessage[], hasMore: boolean }
   */
  getLocalMessage(conversation, startTime = 0, count = 20) {
    const c = getConversationKey(conversation);
    const messages = this.localMessages[c];
    // 直接获取全部
    if (count === 0) {
      return messages || { list: [], hasMore: true };
    }
    // 本地数据中不存在此会话的数据，hasmore标记为true，则要求调用方拉取一次远程历史
    if (!messages) {
      return { list: [], hasMore: true };
    }
    if (startTime === 0) {
      startTime = Date.now();
    }
    let index = messages.list.findIndex((one) => {
      return startTime <= one.sentTime;
    });
    let result = [];
    if (index === 0) {
      // 说明startTime小于第零个的sentTime，则直接用第零个的sentTime开始在服务器查找数据
      return { list: [], hasMore: messages.hasMore };
    } else {
      if (index === -1) {
        // 没有找到，说明starttime大于任何一个messages中的消息sentTime
        index = messages.list.length;
      }
      result = messages.list.slice(Math.max(0, index - count), index);
      if (result.length >= count && index - count > 0) {
        // 通知业务层，本地是否还有更多的历史记录
        // 假设此会话已经将历史记录全部拉取回本地，此时hasMore记录为false（没有更多的历史消息）
        // 当切换回此会话时，下拉加载通知给业务层的会为false，造成判断错误。
        return { list: result, hasMore: true };
      }
      // 本地数据不足，因此通知业务层的hasmore 为服务器返回
      return { list: result, hasMore: messages.hasMore };
    }
  }
  /**
   * 获取远程消息
   * @param conversation
   * @param option
   * @returns Promise<void>
   */
  getRemoteMessages(conversation, option) {
    return getHistoryMessages(conversation, option).then((res) => {
      if (res.code === 0) {
        let list = res.data.list;
        // 排除用户要自己接管的数据
        if (
          this.customIntercept &&
          typeof this.customIntercept.interceptMessage === "function"
        ) {
          list = list.filter((one) => {
            const isMatch = this.customIntercept.interceptMessage(one);
            return !isMatch;
          });
        }
        // 用户处理消息【供客户对消息进行加解密操作】
        if (
          this.customDisplayMessage &&
          this.customDisplayMessage.willDisplayMessages &&
          typeof this.customDisplayMessage.willDisplayMessages === "function"
        ) {
          list = list.map((message) => {
            let content = this.customDisplayMessage.willDisplayMessages(
              message.content,
              message.conversationType,
              message.targetId,
              message.messageType
            );
            message.content = content;
            return message;
          });
        }
        this.setLocalMessages(list, res.data.hasMore);
      }
    });
  }
  /**
   * 按sentTime升序存入
   * @param messages IAReceivedMessage[] 要存入的消息列表
   * @returns
   */
  setLocalMessages(messages = [], hasMore) {
    const result = [];
    if (!messages.length) {
      return { new: result };
    }
    const willRemove = [];
    const c =
      messages[0].conversationType +
      "&&" +
      messages[0].targetId +
      "&&" +
      (messages[0].channelId || "");
    if (this.localMessages[c]) {
      if (hasMore !== undefined) {
        this.localMessages[c].hasMore = hasMore;
      }
    } else {
      // 不存数据的情况下，hasMore设定为true，可以防止丢消息
      this.localMessages[c] = {
        list: [],
        hasMore: hasMore === undefined ? true : hasMore,
      };
    }
    const list = this.localMessages[c].list;
    messages.forEach((message) => {
      let targetPosition = list.length;
      // 收到对方的消息撤回通知，则在本地会话中清理要撤回的消息
      if (
        message.messageType === "RC:RcCmd" &&
        message.messageDirection === MessageDirection.RECEIVE
      ) {
        const content = message.content;
        if (content && content.messageUId) {
          const removeMessage = this.deleteMessageByUId(
            list,
            content.messageUId
          );
          if (removeMessage) {
            willRemove.push(removeMessage);
          }
        }
      }
      for (let i = 0; i < list.length; i++) {
        const localMessage = list[i];
        if (message.messageUId === localMessage.messageUId) {
          // 防止消息重复
          return;
        } else if (message.sentTime <= localMessage.sentTime) {
          targetPosition = i;
          break;
        }
      }
      // 返回当前的会话的更新消息
      if (
        this._currentConversation &&
        isEqual(message, this._currentConversation)
      ) {
        result.push(message);
      }
      // 根据本地存储的已读响应时间戳判断单聊消息的已读未读状态。
      const isPrivat = message.conversationType === ConversationType.PRIVATE;
      const isSendUser = message.messageDirection === MessageDirection.SEND;
      const isUnRead = message.sentTime > this._unReadTime || !this._unReadTime;
      if (isPrivat && isSendUser && isUnRead) {
        // 单聊，自己发送的消息
        message.receivedStatus = ReceivedStatus.UNREAD;
      }
      // 将历史消息写入本地数据
      this.localMessages[c].list.splice(targetPosition, 0, message);
    });
    result.sort((a, b) => {
      return a.sentTime - b.sentTime;
    });
    return { new: result, remove: willRemove };
  }
  /**
   * 撤回消息
   * @param message
   * @returns IAReceivedMessage | null
   */
  recallMessage(message) {
    const key = getConversationKey(message);
    const result = this.localMessages[key];
    if (!result) {
      return null;
    }
    const messages = result.list;
    if (messages && messages.length) {
      const index = messages.findIndex((m) => {
        return m.messageUId === message.messageUId;
      });
      if (index !== -1) {
        // messages[index].messageType += '__RECALL'
        messages[index].messageType = "RC:RcCmd";
        return messages[index];
      }
      return null;
    }
    return null;
  }
  /**
   * 删除消息
   * @param message
   * @returns boolean
   */
  deleteMessage(message) {
    const key = getConversationKey(message);
    const result = this.localMessages[key];
    if (!result) {
      return false;
    }
    const messages = result.list;
    if (messages && messages.length) {
      const index = messages.findIndex((m) => {
        return m.messageUId === message.messageUId;
      });
      if (index !== -1) {
        logger.info(LogTagId.A_KIT_MSG_DELETE_O, {
          index,
          message: messages[index],
        });
        messages.splice(index, 1);
        return true;
      }
      return false;
    }
    return false;
  }
  /**
   * 根据消息UId删除本地消息
   * @param list
   * @param uid
   * @returns
   */
  deleteMessageByUId(list, uid) {
    const index = list.findIndex((one) => {
      return one.messageUId === uid;
    });
    if (index !== -1) {
      const message = list[index];
      list.splice(index, 1);
      logger.info(LogTagId.A_KIT_MSG_DELETE_O, { index, message: list[index] });
      return message;
    }
  }
  /**
   * 插入消息
   * @param message
   * @returns void
   */
  insertMessage(message) {
    const key = getConversationKey(message);
    const result = this.localMessages[key];
    if (!result) {
      this.localMessages[key] = { list: [message], hasMore: true };
      return;
    }
    const list = this.localMessages[key].list;
    const index = list.findIndex((one) => one.sentTime > message.sentTime);
    if (index === -1) {
      list.push(message);
    } else {
      list.splice(index, 0, message);
    }
    logger.info(LogTagId.A_KIT_MSG_INSERT_O, { index, message });
  }
  /**
   * 更新消息
   * @param message
   * @returns void
   */
  updateMessage(message) {
    const key = getConversationKey(message);
    const result = this.localMessages[key];
    if (!result) {
      this.insertMessage(message);
      return;
    }
    const list = this.localMessages[key].list;
    let index = -1;
    // 首先考虑messageId，如果不存在messageId则再用messageUId来
    if (message.messageId) {
      index = list.findIndex((one) => one.messageId === message.messageId);
    }
    if (index === -1 && message.messageUId) {
      index = list.findIndex((one) => one.messageUId === message.messageUId);
    }
    if (index === -1) {
      this.insertMessage(message);
      return;
    }
    for (const p in message) {
      if (message[p] !== undefined) {
        // logger.info(LogTagId.A_KIT_MSG_UPDATE_O, { old: list[index][p], new: message[p] })
        list[index][p] = message[p];
      }
    }
  }
  /**
   * 截取消息，保留最新的100条
   * @returns void
   */
  cutMessages() {
    const key = getConversationKey(this._currentConversation);
    const result = this.localMessages[key];
    if (!result) {
      return;
    }
    if (result.list.length > 100) {
      result.list.splice(0, result.list.length - 100);
      result.hasMore = true;
    }
  }
  /**
   * disconnect 置空本地消息
   */
  clearLocalConversation() {
    this.localMessages = {};
  }
}
